C
C   Primary author:  Hans Johnston  (hansjohn@umich.edu) 
C   Department of Mathematics
C   University of Michigan, Ann Arbor
C
C   Copyright (c) 2003. The Regents of the University of Michigan.
C   All Rights Reserved.
C
C   This program is free software; you can redistribute it and/or modify
C   it under the terms of the GNU General Public License as published by
C   the Free Software Foundation; either version 2 of the License, or
C   (at your option) any later version.
C
C   This program is distributed in the hope that it will be useful,
C   but WITHOUT ANY WARRANTY; without even the implied warranty of
C   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C   GNU General Public License for more details.
C
C   You should have received a copy of the GNU General Public License
C   along with this program; if not, write to the Free Software Foundation,
C   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
C
C
C
C
C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      MODULE treecode3d_procedures
      IMPLICIT NONE

C r8 is 8-byte (double precision) real

      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)

C global variables for taylor expansions

      INTEGER :: torder,torderlim
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: cf,cf1,cf2
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:,:,:) :: b1

C global variables to track tree levels 
 
      INTEGER :: minlevel,maxlevel
      
C global variables used when computing potential/force

      INTEGER :: orderoffset
      REAL(KIND=r8),DIMENSION(3) :: tarpos
      REAL(KIND=r8) :: thetasq

C global variables for postition and charge storage

      INTEGER,ALLOCATABLE,DIMENSION(:)  :: orderarr
      REAL(KIND=r8),ALLOCATABLE,DIMENSION(:) :: xcopy,ycopy,zcopy,qcopy

C node pointer and node type declarations

      TYPE tnode_pointer
           TYPE(tnode), POINTER :: p_to_tnode
      END TYPE tnode_pointer
      TYPE tnode
           INTEGER          :: numpar,ibeg,iend
           REAL(KIND=r8)    :: x_min,y_min,z_min
           REAL(KIND=r8)    :: x_max,y_max,z_max
           REAL(KIND=r8)    :: x_mid,y_mid,z_mid
           REAL(KIND=r8)    :: radius,sqradius,aspect
           INTEGER          :: level,num_children,exist_ms
           REAL(KIND=r8),DIMENSION(:,:,:),POINTER :: ms
           TYPE(tnode_pointer), DIMENSION(8) :: child
      END TYPE tnode

      CONTAINS
!!!!!!!!!!!!!!!
      SUBROUTINE SETUP(x,y,z,q,numpars,order,theta,iflag,xyzminmax,
     &                 arrdim)
      IMPLICIT NONE
C
C SETUP allocates and initializes arrays needed for the Taylor expansion.
C Also, global variables are set and the Cartesian coordinates of
C the smallest box containing the particles is determined. The particle
C postions and charges are copied so that they can be restored upon exit.
C
      INTEGER,INTENT(IN) :: numpars,order,iflag,arrdim
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q
      REAL(KIND=r8),INTENT(INOUT),DIMENSION(6) :: xyzminmax
      REAL(KIND=r8),INTENT(IN) :: theta

C local variables

      INTEGER :: err,i
      REAL(KIND=r8) :: t1

C global integers and reals:  TORDER, TORDERLIM and THETASQ

      torder=order
      IF (iflag .EQ. 1) THEN
          orderoffset=0
      ELSE
          orderoffset=1
      END IF
      torderlim=torder+orderoffset
      thetasq=theta*theta

C allocate global Taylor expansion variables

      ALLOCATE(cf(0:torder),cf1(torder+1),cf2(torder+1),
     &         b1(0:torderlim,0:torderlim,0:torderlim), STAT=err)
      IF (err .NE. 0) THEN
         WRITE(6,*) 'Error allocating Taylor variables! '
         STOP
      END IF

C initialize arrays for Taylor sums and coeffs

      DO i=0,torder
         cf(i)=-(REAL(i,KIND=r8)+1.0_r8)
      END DO
      DO i=1,torderlim
         t1=1.0_r8/REAL(i,KIND=r8)
         cf1(i)=1.0_r8-0.5_r8*t1
         cf2(i)=1.0_r8-t1
      END DO

C find bounds of Cartesian box enclosing the particles

      xyzminmax(1)=MINVAL(x(1:numpars))
      xyzminmax(2)=MAXVAL(x(1:numpars))
      xyzminmax(3)=MINVAL(y(1:numpars))
      xyzminmax(4)=MAXVAL(y(1:numpars))
      xyzminmax(5)=MINVAL(z(1:numpars))
      xyzminmax(6)=MAXVAL(z(1:numpars))


C Allocate arrays and copy variables. Also create and initialize orderarr.

      ALLOCATE(xcopy(numpars),ycopy(numpars),zcopy(numpars), 
     &         qcopy(numpars),orderarr(numpars),STAT=err)
      IF (err .NE. 0) THEN
         WRITE(6,*) 'Error allocating copy variables! '
         STOP
      END IF    

      xcopy=x(1:numpars)
      ycopy=y(1:numpars)
      zcopy=z(1:numpars)
      qcopy=q(1:numpars)

      DO i=1,numpars
         orderarr(i)=i
      END DO  

      RETURN
      END SUBROUTINE SETUP
!!!!!!!!!!!!!
      RECURSIVE SUBROUTINE CREATE_TREE(p,ibeg,iend,x,y,z,q,shrink,
     &                                 maxparnode,xyzmm,level,arrdim)
      IMPLICIT NONE
C
C CREATE_TREE recursively create the tree structure. Node P is
C input, which contains particles indexed from IBEG to IEND. After
C the node parameters are set subdivision occurs if IEND-IBEG+1 > MAXPARNODE.
C Real array XYZMM contains the min and max values of the coordinates
C of the particle in P, thus defining the box.   
C
      TYPE(tnode),POINTER :: p
      INTEGER,INTENT(IN) :: ibeg,iend,shrink,level,maxparnode,arrdim
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(INOUT) :: x,y,z,q
      REAL(KIND=r8),DIMENSION(6),INTENT(IN) :: xyzmm

C local variables

      REAL(KIND=r8) :: x_mid,y_mid,z_mid,xl,yl,zl,lmax,t1,t2,t3
      INTEGER, DIMENSION(8,2) :: ind
      REAL(KIND=r8), DIMENSION(6,8) :: xyzmms
      INTEGER :: i,j,limin,limax,err,loclev,numposchild
      REAL(KIND=r8), DIMENSION(6) ::  lxyzmm
     
C allocate pointer 

      ALLOCATE(p,STAT=err)
      IF (err .NE. 0) THEN
         WRITE(6,*) 'Error allocating pointer! '
         STOP
      END IF

C set node fields: number of particles, exist_ms
C and xyz bounds 

      p%numpar=iend-ibeg+1
      p%exist_ms=0

      IF (shrink .EQ. 1) THEN   
         p%x_min=MINVAL(x(ibeg:iend))
         p%x_max=MAXVAL(x(ibeg:iend))
         p%y_min=MINVAL(y(ibeg:iend))
         p%y_max=MAXVAL(y(ibeg:iend))
         p%z_min=MINVAL(z(ibeg:iend))
         p%z_max=MAXVAL(z(ibeg:iend))
      ELSE
         p%x_min=xyzmm(1)
         p%x_max=xyzmm(2)
         p%y_min=xyzmm(3)
         p%y_max=xyzmm(4)
         p%z_min=xyzmm(5)
         p%z_max=xyzmm(6)        
      END IF

C compute aspect ratio

      xl=p%x_max-p%x_min
      yl=p%y_max-p%y_min
      zl=p%z_max-p%z_min

      lmax=MAX(xl,yl,zl)
      t1=lmax
      t2=MIN(xl,yl,zl)
      IF (t2 .NE. 0.0_r8) THEN
         p%aspect=t1/t2
      ELSE
         p%aspect=0.0_r8
      END IF

C midpoint coordinates , RADIUS and SQRADIUS 

      p%x_mid=(p%x_max+p%x_min)/2.0_r8
      p%y_mid=(p%y_max+p%y_min)/2.0_r8
      p%z_mid=(p%z_max+p%z_min)/2.0_r8
      t1=p%x_max-p%x_mid
      t2=p%y_max-p%y_mid
      t3=p%z_max-p%z_mid
      p%sqradius=t1*t1+t2*t2+t3*t3
      p%radius=SQRT(p%sqradius)

C set particle limits, tree level of node, and nullify children pointers

      p%ibeg=ibeg
      p%iend=iend
      p%level=level
      IF (maxlevel .LT. level) THEN
         maxlevel=level
      END IF
      p%num_children=0
      DO i=1,8
         NULLIFY(p%child(i)%p_to_tnode)
      END DO

      IF (p%numpar .GT. maxparnode) THEN
C
C set IND array to 0 and then call PARTITION routine.  IND array holds indices
C of the eight new subregions. Also, setup XYZMMS array in case SHRINK=1
C
         xyzmms(1,1)=p%x_min
         xyzmms(2,1)=p%x_max
         xyzmms(3,1)=p%y_min
         xyzmms(4,1)=p%y_max
         xyzmms(5,1)=p%z_min
         xyzmms(6,1)=p%z_max
         ind(1,1)=ibeg
         ind(1,2)=iend
         x_mid=p%x_mid
         y_mid=p%y_mid
         z_mid=p%z_mid

         CALL PARTITION_8(x,y,z,q,xyzmms,xl,yl,zl,lmax,numposchild,
     &                    x_mid,y_mid,z_mid,ind,arrdim)
C
C create children if indicated and store info in parent
C
         loclev=level+1
         DO i=1,numposchild
            IF (ind(i,1) .LE. ind(i,2)) THEN
               p%num_children=p%num_children+1
               lxyzmm=xyzmms(:,i)
               CALL CREATE_TREE(p%child(p%num_children)%p_to_tnode,
     &                          ind(i,1),ind(i,2),x,y,z,q,shrink,
     &                          maxparnode,lxyzmm,loclev,arrdim)
            END IF
         END DO
      ELSE
         IF (level .LT. minlevel) THEN
            minlevel=level
         END IF
      END IF   

      END SUBROUTINE CREATE_TREE      
!!!!!!!!!!!!!!!
      SUBROUTINE PARTITION_8(x,y,z,q,xyzmms,xl,yl,zl,lmax,numposchild,
     &                       x_mid,y_mid,z_mid,ind,arrdim)
      IMPLICIT NONE
C
C PARTITION_8 determines the particle indices of the eight sub boxes
C containing the particles after the box defined by particles I_BEG
C to I_END is divided by its midpoints in each coordinate direction.
C The determination of the indices is accomplished by the subroutine
C PARTITION. A box is divided in a coordinate direction as long as the
C resulting aspect ratio is not too large. This avoids the creation of
C "narrow" boxes in which Talyor expansions may become inefficient.
C On exit the INTEGER array IND (dimension 8 x 2) contains
C the indice limits of each new box (node) and NUMPOSCHILD the number 
C of possible children.  If IND(J,1) > IND(J,2) for a given J this indicates
C that box J is empty.
C
      INTEGER, INTENT(IN) :: arrdim
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(INOUT) :: x,y,z,q
      INTEGER, DIMENSION(8,2),INTENT(INOUT) :: ind
      REAL(KIND=r8),DIMENSION(6,8),INTENT(INOUT) :: xyzmms
      REAL(KIND=r8), INTENT(IN) :: x_mid,y_mid,z_mid,xl,yl,zl,lmax
      INTEGER,INTENT(INOUT) :: numposchild

C local variables

      INTEGER :: temp_ind,i
      REAL(KIND=r8) :: critlen

      numposchild=1
      critlen=lmax/sqrt(2.0_r8)

      IF (xl .GE. critlen) THEN
         CALL PARTITION(x,y,z,q,orderarr,ind(1,1),ind(1,2),
     &                  x_mid,temp_ind,arrdim)

         ind(2,1)=temp_ind+1
         ind(2,2)=ind(1,2)
         ind(1,2)=temp_ind
         xyzmms(:,2)=xyzmms(:,1)
         xyzmms(2,1)=x_mid
         xyzmms(1,2)=x_mid
         numposchild=2*numposchild
      END IF 
 
      IF (yl .GE. critlen) THEN
         DO i=1,numposchild
            CALL PARTITION(y,x,z,q,orderarr,ind(i,1),ind(i,2),
     &                     y_mid,temp_ind,arrdim)
            ind(numposchild+i,1)=temp_ind+1
            ind(numposchild+i,2)=ind(i,2)
            ind(i,2)=temp_ind
            xyzmms(:,numposchild+i)=xyzmms(:,i)
            xyzmms(4,i)=y_mid
            xyzmms(3,numposchild+i)=y_mid
         END DO
         numposchild=2*numposchild
      END IF

      IF (zl .GE. critlen) THEN
         DO i=1,numposchild
            CALL PARTITION(z,x,y,q,orderarr,ind(i,1),ind(i,2),
     &                     z_mid,temp_ind,arrdim)
            ind(numposchild+i,1)=temp_ind+1
            ind(numposchild+i,2)=ind(i,2)
            ind(i,2)=temp_ind
            xyzmms(:,numposchild+i)=xyzmms(:,i)
            xyzmms(6,i)=z_mid
            xyzmms(5,numposchild+i)=z_mid
         END DO
         numposchild=2*numposchild
      END IF

      RETURN 
      END SUBROUTINE PARTITION_8
!!!!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE TREE_COMPFP(p,x,y,z,q,tpeng,fx,fy,fz,numpars,
     &                       farrdim,arrdim)
      IMPLICIT NONE
C
C TREE_COMPF is the driver routine which calls COMPF_TREE for each
C particle, setting the global variable TARPOS before the call. 
C The current target particle's x coordinate and charge are changed
C so that it does not interact with itself. P is the root node of the tree. 
C

      INTEGER,INTENT(IN) :: numpars,farrdim,arrdim
      TYPE(tnode),POINTER :: p  
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(INOUT) :: x,q
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: y,z
      REAL(KIND=r8),DIMENSION(farrdim),INTENT(INOUT) :: fx,fy,fz
      REAL(KIND=r8),INTENT(INOUT) :: tpeng

C local variables

      INTEGER :: i,j
      REAL(KIND=r8),DIMENSION(3) :: flocal,f
      REAL(KIND=r8) :: penglocal,peng,tempx,tempq


      tpeng=0.0_r8
      DO i=1,numpars
         f=0.0_r8
         peng=0.0_r8
         tarpos(1)=x(i)
         tarpos(2)=y(i)
         tarpos(3)=z(i)
         tempx=x(i)
         tempq=q(i)
         x(i)=x(i)+100.123456789_r8
         q(i)=0.0_r8
         DO j=1,p%num_children
            CALL COMPFP_TREE(p%child(j)%p_to_tnode,penglocal,
     &                      flocal,x,y,z,q,arrdim)
            f=f+flocal
            peng=peng+penglocal     
         END DO
         x(i)=tempx
         q(i)=tempq
         fx(orderarr(i))=-tempq*f(1)
         fy(orderarr(i))=-tempq*f(2)
         fz(orderarr(i))=-tempq*f(3)
         tpeng=tpeng+tempq*peng
      END DO
      tpeng=0.5_r8*tpeng

      RETURN
      END SUBROUTINE TREE_COMPFP
!!!!!!!!!!!!!!
      RECURSIVE SUBROUTINE COMPFP_TREE(p,peng,f,x,y,z,q,arrdim)
      IMPLICIT NONE

      INTEGER,INTENT(IN) :: arrdim
      TYPE(tnode),POINTER :: p      
      REAL(KIND=r8),INTENT(INOUT) :: peng
      REAL(KIND=r8),DIMENSION(3),INTENT(INOUT) :: f
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q

C local variables

      REAL(KIND=r8),DIMENSION(3) :: flocal
      REAL(KIND=r8) :: tx,ty,tz,distsq,t1,t2,penglocal,cfj,cfk
      INTEGER :: i,j,k,err

C determine DISTSQ for MAC test

      tx=tarpos(1)-p%x_mid
      ty=tarpos(2)-p%y_mid
      tz=tarpos(3)-p%z_mid
      distsq=tx*tx+ty*ty+tz*tz

C intialize potential energy and force 

      peng=0.0_r8
      f=0.0_r8

C If MAC is accepted and there is more than 1 particle in the 
C box use the expansion for the approximation.

      IF ((p%sqradius .LT. distsq*thetasq) .AND.
     &    (p%sqradius .NE. 0.0_r8)) THEN
         IF (p%exist_ms .EQ. 0) THEN
             ALLOCATE(p%ms(0:torder,0:torder,0:torder),STAT=err)
             IF (err .NE. 0) THEN
                WRITE(6,*) 'Error allocating node moments! '
                STOP
             END IF
             CALL COMP_MS(p,x,y,z,q,arrdim)
             p%exist_ms=1
         END IF   
         penglocal=0.0_r8   
         flocal=0.0_r8
         CALL COMP_TCOEFF(p)
         DO k=0,torder
            cfk=cf(k)
            DO j=0,torder-k
               cfj=cf(j)
               DO i=0,torder-k-j
                  flocal(1)=flocal(1)+cf(i)*b1(i+1,j,k)
     &                     *p%ms(i,j,k)
                  flocal(2)=flocal(2)+cfj*b1(i,j+1,k)
     &                     *p%ms(i,j,k)
                  flocal(3)=flocal(3)+cfk*b1(i,j,k+1)
     &                     *p%ms(i,j,k)
                  peng=peng+b1(i,j,k)*p%ms(i,j,k)
               END DO
            END DO
         END DO
         f=f+flocal

      ELSE

C If MAC fails check to see if the are children. If not, perform direct 
C calculation.  If there are children, call routine recursively for each.
C
         IF (p%num_children .EQ. 0) THEN
            CALL COMPFP_DIRECT(penglocal,flocal,p%ibeg,p%iend,
     &                         x,y,z,q,arrdim)
            peng=penglocal
            f=flocal
         ELSE
            DO i=1,p%num_children
               CALL COMPFP_TREE(p%child(i)%p_to_tnode,penglocal,
     &                         flocal,x,y,z,q,arrdim)
               peng=peng+penglocal
               f=f+flocal
            END DO  
         END IF 
      END IF

      RETURN
      END SUBROUTINE COMPFP_TREE
!!!!!!!!
      SUBROUTINE COMP_TCOEFF(p)
      IMPLICIT NONE
C
C COMP_TCOEFF computes the Taylor coefficients of the potential
C using a recurrence formula.  The center of the expansion is the
C midpoint of the node P.  TARPOS and TORDERLIM are globally defined.  
C
      TYPE(tnode),POINTER :: p      

C local varaibles

      REAL(KIND=r8) :: dx,dy,dz,tdx,tdy,tdz,fac,sqfac,t1
      INTEGER :: i,j,k,tp1

C setup variables

      dx=tarpos(1)-p%x_mid
      dy=tarpos(2)-p%y_mid
      dz=tarpos(3)-p%z_mid
      tdx=2.0_r8*dx
      tdy=2.0_r8*dy
      tdz=2.0_r8*dz
      fac=1.0_r8/(dx*dx+dy*dy+dz*dz)
      sqfac=SQRT(fac)

C 0th coeff or function val 

      b1(0,0,0)=sqfac

C 2 indices are 0

      b1(1,0,0)=fac*dx*sqfac
      b1(0,1,0)=fac*dy*sqfac
      b1(0,0,1)=fac*dz*sqfac
  
      DO i=2,torderlim
         b1(i,0,0)=fac*(tdx*cf1(i)*b1(i-1,0,0)-cf2(i)*b1(i-2,0,0))
         b1(0,i,0)=fac*(tdy*cf1(i)*b1(0,i-1,0)-cf2(i)*b1(0,i-2,0))
         b1(0,0,i)=fac*(tdz*cf1(i)*b1(0,0,i-1)-cf2(i)*b1(0,0,i-2))
      END DO

C 1 index 0, 1 index 1, other >=1

      b1(1,1,0)=fac*(dx*b1(0,1,0)+tdy*b1(1,0,0))
      b1(1,0,1)=fac*(dx*b1(0,0,1)+tdz*b1(1,0,0))
      b1(0,1,1)=fac*(dy*b1(0,0,1)+tdz*b1(0,1,0))

      DO i=2,torderlim-1
         b1(1,0,i)=fac*(dx*b1(0,0,i)+tdz*b1(1,0,i-1)-b1(1,0,i-2))
         b1(0,1,i)=fac*(dy*b1(0,0,i)+tdz*b1(0,1,i-1)-b1(0,1,i-2))
         b1(0,i,1)=fac*(dz*b1(0,i,0)+tdy*b1(0,i-1,1)-b1(0,i-2,1))
         b1(1,i,0)=fac*(dx*b1(0,i,0)+tdy*b1(1,i-1,0)-b1(1,i-2,0))
         b1(i,1,0)=fac*(dy*b1(i,0,0)+tdx*b1(i-1,1,0)-b1(i-2,1,0))
         b1(i,0,1)=fac*(dz*b1(i,0,0)+tdx*b1(i-1,0,1)-b1(i-2,0,1))
      END DO
  
C 1 index 0, others >= 2
      
      DO i=2,torderlim-2 
         DO j=2,torderlim-i
            b1(i,j,0)=fac*(tdx*cf1(i)*b1(i-1,j,0)+tdy*b1(i,j-1,0)
     &                       -cf2(i)*b1(i-2,j,0)-b1(i,j-2,0))
            b1(i,0,j)=fac*(tdx*cf1(i)*b1(i-1,0,j)+tdz*b1(i,0,j-1)
     &                       -cf2(i)*b1(i-2,0,j)-b1(i,0,j-2))
            b1(0,i,j)=fac*(tdy*cf1(i)*b1(0,i-1,j)+tdz*b1(0,i,j-1)
     &                       -cf2(i)*b1(0,i-2,j)-b1(0,i,j-2))   
         END DO
      END DO

 
C 2 indices 1, other >= 1
C b(1,1,1) is correct, but a little tricky!  
C      b(1,1,1)=5.0*dz*fac*b(1,1,0)

      b1(1,1,1)=fac*(dx*b1(0,1,1)+tdy*b1(1,0,1)+tdz*b1(1,1,0))
      DO i=2,torderlim-2
         b1(1,1,i)=fac*(dx*b1(0,1,i)+tdy*b1(1,0,i)+tdz*b1(1,1,i-1)
     &                   -b1(1,1,i-2))
         b1(1,i,1)=fac*(dx*b1(0,i,1)+tdy*b1(1,i-1,1)+tdz*b1(1,i,0)
     &                   -b1(1,i-2,1))
         b1(i,1,1)=fac*(dy*b1(i,0,1)+tdx*b1(i-1,1,1)+tdz*b1(i,1,0)
     &                   -b1(i-2,1,1))
      END DO

C 1 index 1, others >=2
      DO i=2,torderlim-3
         DO j=2,torderlim-i
            b1(1,i,j)=fac*(dx*b1(0,i,j)+tdy*b1(1,i-1,j)+tdz*b1(1,i,j-1)
     &                      -b1(1,i-2,j)-b1(1,i,j-2))
            b1(i,1,j)=fac*(dy*b1(i,0,j)+tdx*b1(i-1,1,j)+tdz*b1(i,1,j-1)
     &                      -b1(i-2,1,j)-b1(i,1,j-2))
            b1(i,j,1)=fac*(dz*b1(i,j,0)+tdx*b1(i-1,j,1)+tdy*b1(i,j-1,1)
     &                      -b1(i-2,j,1)-b1(i,j-2,1))
         END DO
      END DO

C all indices >=2
      DO k=2,torderlim-4
         DO j=2,torderlim-2-k
            DO i=2,torderlim-k-j
               b1(i,j,k)=fac*(tdx*cf1(i)*b1(i-1,j,k)+tdy*b1(i,j-1,k)
     &                      +tdz*b1(i,j,k-1)-cf2(i)*b1(i-2,j,k)
     &                      -b1(i,j-2,k)-b1(i,j,k-2)) 
            END DO
         END DO
      END DO

      RETURN
      END SUBROUTINE COMP_TCOEFF    
!!!!!!!!!!!!!!!
      SUBROUTINE COMP_MS(p,x,y,z,q,arrdim)
      IMPLICIT NONE
C
C COMP_MS computes the moments for node P needed in the Taylor approximation
C
      INTEGER,INTENT(IN) :: arrdim
      TYPE(tnode),POINTER :: p 
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q

C local variables

      INTEGER :: i,k1,k2,k3
      REAL(KIND=r8) :: dx,dy,dz,tx,ty,tz
     
      p%ms=0.0_r8
      DO i=p%ibeg,p%iend
         dx=x(i)-p%x_mid
         dy=y(i)-p%y_mid
         dz=z(i)-p%z_mid
         tz=1.0_r8
         DO k3=0,torder
            ty=1.0_r8
            DO k2=0,torder-k3
               tx=1.0_r8 
               DO k1=0,torder-k3-k2
                  p%ms(k1,k2,k3)=p%ms(k1,k2,k3)+q(i)*tx*ty*tz
                  tx=tx*dx
              END DO
              ty=ty*dy
            END DO
            tz=tz*dz
         END DO
      END DO
         
      RETURN
      END SUBROUTINE COMP_MS
!!!!!!!!!!!!!!!!
      SUBROUTINE COMPFP_DIRECT(peng,f,ibeg,iend,x,y,z,q,arrdim)
      IMPLICIT NONE
C
C COMPF_DIRECT directly computes the force on the current target
C particle determined by the global variable TARPOS. 
C
      INTEGER,INTENT(IN) :: ibeg,iend,arrdim
      REAL(KIND=r8),DIMENSION(3),INTENT(OUT) :: f
      REAL(KIND=r8),INTENT(OUT) :: peng
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q

C local variables

      INTEGER :: i
      REAL(KIND=r8) :: dist,t1,tx,ty,tz

      peng=0.0_r8
      f=0.0_r8
      DO i=ibeg,iend
         tx=x(i)-tarpos(1)
         ty=y(i)-tarpos(2)
         tz=z(i)-tarpos(3)
         dist=1.0_r8/SQRT(tx*tx+ty*ty+tz*tz)
         t1=dist*dist*dist
         peng=peng+q(i)*dist
         f(1)=f(1)+q(i)*tx*t1
         f(2)=f(2)+q(i)*ty*t1
         f(3)=f(3)+q(i)*tz*t1 
      END DO   

      RETURN
      END SUBROUTINE COMPFP_DIRECT
!!!!!!!!!!!!!!!!!!!!!!!!
      SUBROUTINE TREE_COMPP(p,x,y,z,q,tpeng,numpars,arrdim)
      IMPLICIT NONE
C
C TREE_COMPF is the driver routine which calls COMPP_TREE for each
C particle, setting the global variables TARPOS before the call.
C The current target particle's x coordinate and charge are changed
C so that it does not interact with itself. P is the root node of the tree. 
C
      INTEGER,INTENT(IN) :: numpars,arrdim
      TYPE(tnode),POINTER :: p  
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(INOUT) :: x,q
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: y,z
      REAL(KIND=r8),INTENT(INOUT) :: tpeng

C local variables

      INTEGER :: i,j
      REAL(KIND=r8) :: penglocal,peng,tempx,tempq

      tpeng=0.0_r8
      DO i=1,numpars
         peng=0.0_r8
         tarpos(1)=x(i)
         tarpos(2)=y(i)
         tarpos(3)=z(i)
         tempx=x(i)
         tempq=q(i)
         x(i)=x(i)+100.123456789_r8
         q(i)=0.0_r8
         DO j=1,p%num_children
            CALL COMPP_TREE(p%child(j)%p_to_tnode,penglocal,
     &                       x,y,z,q,arrdim)
            peng=peng+penglocal     
         END DO
         x(i)=tempx
         q(i)=tempq
         tpeng=tpeng+tempq*peng
      END DO
      tpeng=0.5_r8*tpeng

      RETURN
      END SUBROUTINE TREE_COMPP
!!!!!!!!!!!!!!
      RECURSIVE SUBROUTINE COMPP_TREE(p,peng,x,y,z,q,arrdim)
      IMPLICIT NONE

      INTEGER,INTENT(IN) :: arrdim
      TYPE(tnode),POINTER :: p      
      REAL(KIND=r8),INTENT(INOUT) :: peng
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q

C local variables

      REAL(KIND=r8) :: tx,ty,tz,distsq,t1,t2,penglocal
      INTEGER :: i,j,k,err

C determine DISTSQ for MAC test

      tx=tarpos(1)-p%x_mid
      ty=tarpos(2)-p%y_mid
      tz=tarpos(3)-p%z_mid
      distsq=tx*tx+ty*ty+tz*tz

C intialize potential energy 

      peng=0.0_r8

C If MAC is accepted and there is more than 1 particle in the 
C box use the expansion for the approximation.

      IF ((p%sqradius .LT. distsq*thetasq) .AND.
     &    (p%sqradius .NE. 0.0_r8)) THEN
         IF (p%exist_ms .EQ. 0) THEN
             ALLOCATE(p%ms(0:torder,0:torder,0:torder),STAT=err)
             IF (err .NE. 0) THEN
                WRITE(6,*) 'Error allocating node moments! '
                STOP
             END IF
             CALL COMP_MS(p,x,y,z,q,arrdim)
             p%exist_ms=1
         END IF   
         penglocal=0.0_r8   
         CALL COMP_TCOEFF(p)
         DO k=0,torder
            DO j=0,torder-k
               DO i=0,torder-k-j
                  peng=peng+b1(i,j,k)*p%ms(i,j,k)
               END DO
            END DO
         END DO

      ELSE

C If MAC fails check to see if the are children. If not, perform
C direct calculation.  If there are children, call routine recursively
C for each.

         IF (p%num_children .EQ. 0) THEN
            CALL COMPP_DIRECT(penglocal,p%ibeg,p%iend,
     &                         x,y,z,q,arrdim)
            peng=penglocal
         ELSE
            DO i=1,p%num_children
               CALL COMPP_TREE(p%child(i)%p_to_tnode,penglocal,
     &                         x,y,z,q,arrdim)
               peng=peng+penglocal
            END DO  
         END IF 
      END IF

      RETURN
      END SUBROUTINE COMPP_TREE
!!!!!!!!!!!!!!
      SUBROUTINE COMPP_DIRECT(peng,ibeg,iend,x,y,z,q,arrdim)
      IMPLICIT NONE
C
C COMPP_DIRECT directly computes the potential on the current target
C particle (determined by the global variable TARPOS). 
C
      INTEGER,INTENT(IN) :: ibeg,iend,arrdim
      REAL(KIND=r8),INTENT(OUT) :: peng
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(IN) :: x,y,z,q

C local variables

      INTEGER :: i
      REAL(KIND=r8) :: t1,tx,ty,tz

      peng=0.0_r8
      DO i=ibeg,iend
         tx=x(i)-tarpos(1)
         ty=y(i)-tarpos(2)
         tz=z(i)-tarpos(3)
         peng=peng+q(i)/SQRT(tx*tx+ty*ty+tz*tz)
      END DO   

      RETURN
      END SUBROUTINE COMPP_DIRECT
!!!!!!!!!!!!!!!!!
      SUBROUTINE CLEANUP(p)
      IMPLICIT NONE
C
C CLEANUP deallocates allocated global variables and then
C calls recursive routine REMOVE_NODE to delete the tree.
C
      TYPE(tnode),POINTER :: p      

C local variables
  
      INTEGER :: err

      DEALLOCATE(cf,cf1,cf2,b1, STAT=err)
      IF (err .NE. 0) THEN
         WRITE(6,*) 'Error deallocating Taylor variables! '
         STOP
      END IF      

      DEALLOCATE(xcopy,ycopy,zcopy,qcopy,orderarr,STAT=err)
      IF (err .NE. 0) THEN
         WRITE(6,*) 'Error deallocating copy variables! '
         STOP
      END IF  

      CALL REMOVE_NODE(p)
      DEALLOCATE(p, STAT=err)
      IF (err .NE. 0) THEN
         WRITE(6,*) 'Error deallocating root node! '
         STOP
      END IF 
      NULLIFY(p)         

      RETURN
      END SUBROUTINE CLEANUP
!!!!!!!!!!!
      RECURSIVE SUBROUTINE REMOVE_NODE(p)
      IMPLICIT NONE
C
C REMOVE_NODE recursively removes each node from the
C tree and deallocates its memory for MS array if it
C exits.
C
      TYPE(tnode),POINTER :: p 

C local variables

      INTEGER :: i,err

      IF (p%exist_ms .EQ. 1) THEN
         DEALLOCATE(p%ms,STAT=err)
         IF (err .NE. 0) THEN
            WRITE(6,*) 'Error deallocating node MS! '
            STOP
         END IF               
      END IF

      IF (p%num_children .GT. 0) THEN
          DO i=1,p%num_children
            CALL REMOVE_NODE(p%child(i)%p_to_tnode)
            DEALLOCATE(p%child(i)%p_to_tnode,STAT=err)
            IF (err .NE. 0) THEN
               WRITE(6,*) 'Error deallocating node child! '
               STOP
            END IF                           
          END DO
      END IF 

      RETURN                
      END SUBROUTINE REMOVE_NODE      

      END MODULE treecode3d_procedures
!!!!!!!!!!!!!!!
      SUBROUTINE TREECODE3D(x,y,z,q,numpars,fx,fy,fz,tpeng,iflag,order,
     &                       theta,shrink,maxparnode,timetree,prinfo,
     &                       farrdim,arrdim)
      USE treecode3d_procedures
      IMPLICIT NONE
C
C TREECODE3D approximates the electrostatic force on particles using a hierarchical
C oct-tree algorithm with body-cell forces approximated via Taylor expansions. 
C
C Input:
C
C X,Y,Z : (REAL one-dimensional arrays of length ARRDIM) Cartesian coordinates
C         of the particles 
C
C Q   :  (REAL one-dimensional array of length ARRDIM) Corresponding charge 
C         of each particle
C
C NUMPARS : (INTEGER)  number of particles
C
C IFLAG   : (INTEGER)  flag indicating what is to be computer:
C                  1 - potential energy, 2 - potential energy and force
C
C ORDER :  (INTEGER) order of Taylor expansion used in the body-cell approximation
C
C THETA : (REAL) opening angle used in the BMAX MAC 
C
C SHRINK : (INTEGER) switch used in the oct-tree construction. If equal to 1
C          then upon subdivision of a node the node bounds are taken
C          the smallest Cartesian box containing the particles in the node.
C
C MAXPARNODE : (INTEGER) maximum number of particles allowed in a leaf.
C
C PRINFO : (INTEGER) swith for outputting information in this routine.  If set
C          to 1 then information is writen to stdout.
C
C FARRDIM : (INTEGER) declared length of arrays FX, FY, and FZ. 
C
C ARRDIM  : (INTEGER) declared length of arrays X, Y, Z, and Q.
C
C Output:
C
C FX,FY,FZ : (REAL one-dimensional arrays of length FARRDIM) force on ith particle
C            is given by (FX(i),FY(i),FZ(i)).
C
C TPENG    : (REAL) potential energy.
C
C TREETIME : (REAL) time (in seconds) for the tree algorithm computation.
C
C
      INTEGER,INTENT(IN) :: numpars,iflag,order,farrdim,arrdim,
     &                      shrink,prinfo,maxparnode
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(INOUT) :: x,y,z,q
      REAL(KIND=r8),DIMENSION(farrdim),INTENT(INOUT) :: fx,fy,fz
      REAL(KIND=r8),INTENT(IN) :: theta
      REAL(KIND=r8),INTENT(OUT) :: tpeng,timetree

C local variables

      TYPE(tnode),POINTER :: troot
      INTEGER :: i,level,err
      REAL(KIND=r8), DIMENSION(6) :: xyzminmax

C variables needed for f90 DATE_AND_TIME intrinsic

      INTEGER,DIMENSION(8) :: time1,time2 
      CHARACTER (LEN=8)  :: datec
      CHARACTER (LEN=10) :: timec
      CHARACTER (LEN=5)  :: zonec
      REAL(KIND=r8)      :: totaltime


C Call SETUP to allocate arrays for Taylor expansions
C and setup global variables. Also, copy variables into global copy arrays. 

      CALL SETUP(x,y,z,q,numpars,order,theta,iflag,xyzminmax,arrdim)

C nullify pointer to root of tree (TROOT) and create tree

      NULLIFY(troot)  

      CALL DATE_AND_TIME(datec,timec,zonec,time1)

C set global variables to track tree levels during construction

      level=0
      minlevel=50000
      maxlevel=0

      IF (prinfo .EQ. 1) THEN
         WRITE(6,*) ' '
         WRITE(6,*) 'Creating tree '
      END IF

      CALL CREATE_TREE(troot,1,numpars,x,y,z,q,shrink,
     &                 maxparnode,xyzminmax,level,arrdim)

      CALL DATE_AND_TIME(datec,timec,zonec,time2)
      CALL TTIME(time1,time2,totaltime)

C print tree information to stdout if PRINFO=1

      IF (prinfo .EQ. 1) THEN

         WRITE(6,*) ' '
         WRITE(6,*) 'Tree created '
         WRITE(6,*) ' '
         WRITE(6,*) 'Tree creation time (secs) : ',totaltime
         WRITE(6,*) ' '
         WRITE(6,*) 'Run synopsis : '
         WRITE(6,*) ' '
         WRITE(6,*) 'numpar :',troot%numpar
         WRITE(6,*) 'x_mid  :',troot%x_mid
         WRITE(6,*) 'y_mid  :',troot%y_mid
         WRITE(6,*) 'z_mid  :',troot%z_mid
         WRITE(6,*) 'radius :',troot%radius   
         WRITE(6,*) 'torder          :',torder
         WRITE(6,*) 'theta           :',theta
         WRITE(6,*) 'shrink          :',shrink
         WRITE(6,*) 'maxparnode      :',maxparnode
         WRITE(6,*) 'iflag           :',iflag
         WRITE(6,*) 'tree minlevel   :',minlevel
         WRITE(6,*) 'tree maxlevel   :',maxlevel
         WRITE(6,*) ' '
      END IF
 
      CALL DATE_AND_TIME(datec,timec,zonec,time1)

      IF (iflag .NE. 1) THEN
         IF (prinfo .EQ. 1) THEN
            WRITE(6,*) ' '
            WRITE(6,*) 'Computing potential energy and force with tree '
         END IF
         CALL TREE_COMPFP(troot,x,y,z,q,tpeng,fx,fy,fz,numpars,
     &                    farrdim,arrdim)
      ELSE
         IF (prinfo .EQ. 1) THEN
            WRITE(6,*) ' '
            WRITE(6,*) 'Computing potential energy with tree '
         END IF
         CALL TREE_COMPP(troot,x,y,z,q,tpeng,numpars,arrdim)
      END IF

      CALL DATE_AND_TIME(datec,timec,zonec,time2)
      CALL TTIME(time1,time2,totaltime)
      timetree=totaltime

      IF (prinfo .EQ. 1) THEN      
          WRITE(6,*) ' '
          WRITE(6,*) 'Treecode time (secs) : ',timetree
      END IF

C restore x, y, z and q to their original values

      x(1:numpars)=xcopy
      y(1:numpars)=ycopy
      z(1:numpars)=zcopy
      q(1:numpars)=qcopy

C Call CLEANUP to deallocate global variables and tree structure.

      IF (prinfo .EQ. 1) THEN
         WRITE(6,*) ' '
         WRITE(6,*) 'Deallocating tree structure!'
         WRITE(6,*) ' '
      END IF

      CALL CLEANUP(troot)

      END SUBROUTINE TREECODE3D
!!!!!!!!!!!!!
      SUBROUTINE PARTITION(a,b,c,q,indarr,ibeg,iend,val,midind,arrdim)
      IMPLICIT NONE
C
C PARTITION determines the index MIDIND, after partitioning
C in place the  arrays A,B,C and Q,  such that 
C A(IBEG:MIDIND) <= VAL and  A(MIDIND+1:IEND) > VAL. 
C If on entry IBEG > IEND or  A(IBEG:IEND) > VAL then MIDIND
C is returned as IBEG-1. 
C 
      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)
      INTEGER, INTENT(IN) :: arrdim,ibeg,iend
      REAL(KIND=r8),DIMENSION(arrdim),INTENT(INOUT) :: a,b,c,q
      INTEGER,DIMENSION(arrdim),INTENT(INOUT) :: indarr   
      INTEGER, INTENT(INOUT) :: midind   
      REAL(KIND=r8) val

C local variables

      REAL(KIND=r8) ta,tb,tc,tq
      INTEGER lower,upper,tind

      IF (ibeg .LT. iend) THEN

C temporarily store IBEG entries and set A(IBEG)=VAL for 
C the partitoning algorithm.  

         ta=a(ibeg)
         tb=b(ibeg)
         tc=c(ibeg)
         tq=q(ibeg)
         tind=indarr(ibeg)
         a(ibeg)=val 
         upper=ibeg
         lower=iend

         DO WHILE (upper .NE. lower)
            DO WHILE ((upper .LT. lower) .AND. (val .LT. a(lower)))
                  lower=lower-1
            END DO
            IF (upper .NE. lower) THEN
               a(upper)=a(lower)
               b(upper)=b(lower)
               c(upper)=c(lower)
               q(upper)=q(lower)
               indarr(upper)=indarr(lower)
            END IF
            DO WHILE ((upper .LT. lower) .AND. (val .GE. a(upper)))
                  upper=upper+1
            END DO
            IF (upper .NE. lower) THEN
               a(lower)=a(upper)
               b(lower)=b(upper)
               c(lower)=c(upper)
               q(lower)=q(upper)
               indarr(lower)=indarr(upper)
            END IF
         END DO
         midind=upper

C replace TA in position UPPER and change MIDIND if TA > VAL 

         IF (ta .GT. val) THEN
            midind=upper-1
         END IF
         a(upper)=ta
         b(upper)=tb
         c(upper)=tc
         q(upper)=tq
         indarr(upper)=tind

      ELSEIF (ibeg .EQ. iend) THEN
         IF (a(ibeg) .LE. val) THEN
            midind=ibeg
         ELSE
            midind=ibeg-1
         END IF
      ELSE
         midind=ibeg-1
      END IF

      RETURN
      END SUBROUTINE PARTITION
!!!!!!!!!!!!!!!!!!
      SUBROUTINE TTIME(timebeg,timeend,totaltime)
      IMPLICIT NONE
C
C TTIME computes the time difference in seconds between
C the timestamps TIMEBEG and TIMEEND returned by the 
C f90 intrinsic DATE_AND_TIME
C
      INTEGER,PARAMETER :: r8=SELECTED_REAL_KIND(12)
      INTEGER,DIMENSION(8),INTENT(INOUT) :: timebeg,timeend
      REAL(KIND=r8),INTENT(OUT) :: totaltime

C TIMEEND is modifed by borrowing in case each of its fields
C are not .GE. to the corresponding field in TIMEBEG (up to
C and including days) 

      IF (timeend(8) .LT. timebeg(8)) THEN
          timeend(8)=timeend(8)+1000
          timeend(7)=timeend(7)-1
      END IF
      IF (timeend(7) .LT. timebeg(7)) THEN
          timeend(7)=timeend(7)+60
          timeend(6)=timeend(6)-1
      END IF
      IF (timeend(6) .LT. timebeg(6)) THEN
          timeend(6)=timeend(6)+60
          timeend(5)=timeend(5)-1
      END IF
      IF (timeend(5) .LT. timebeg(5)) THEN
          timeend(5)=timeend(5)+24
          timeend(3)=timeend(3)-1
      END IF

      totaltime=  REAL(timeend(8)-timebeg(8),KIND=r8) +
     &       1000.0_r8*( REAL(timeend(7)-timebeg(7),KIND=r8) + 
     &         60.0_r8*( REAL(timeend(6)-timebeg(6),KIND=r8) +
     &         60.0_r8*( REAL(timeend(5)-timebeg(5),KIND=r8) +
     &         24.0_r8*( REAL(timeend(3)-timebeg(3),KIND=r8)))))
      totaltime=totaltime/1000.0_r8

     
      RETURN
      END SUBROUTINE TTIME



